# -*- coding: utf-8 -*-
# Copyright (C) 2020 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.modules.quorum"""

import re
import logging

from copy import deepcopy

import requests
from six import iteritems

from dwho.classes.modules import DWhoModuleBase, MODULES
from httpdis.ext.httpdis_json import HttpReqErrJson
from sonicprobe import helpers
from sonicprobe.libs import network, urisup, xys
from sonicprobe.libs.moresynchro import RWLock

from ..classes.common import DB_REPLICATION_TYPE_DEFAULT, DB_REPLICATION_TYPE_OPTS
from ..classes.abstract.db import XtrargonDBBase
from ..classes.ec2 import XtrargonEc2Helper

LOG                             = logging.getLogger('xtrargon_ng.modules.quorum')

DB_NODE_LETTERS                 = {'a': 'b',
                                   'b': 'a'}

DEFAULT_CNF_FILE                = "quorum.cnf"

DEFAULT_BACKUPS_STATUS_URI      = "http://:0/backups/status"
DEFAULT_FORBID_VIP_URI          = "http://:0/server/forbid-vip"
DEFAULT_REQUEST_TIMEOUT         = 15
DEFAULT_FREEZE_VIP_FILE_PATTERN = "/var/opt/vip/%s_freeze"

MATCH_DB_NODE                   = re.compile(r'^([a-z0-9][a-z0-9\-\.]+[a-z0-9]-[0-9]+)([a-z])(\.[a-z0-9\-\.]+[a-z0-9]|)$').match

xys.add_callback('quorum.host', network.valid_domain)

class QuorumDB(XtrargonDBBase): # pylint: disable-msg=too-few-public-methods
    pass

class QuorumModule(DWhoModuleBase):
    MODULE_NAME = 'quorum'

    LOCK = RWLock()

    @staticmethod
    def _load_ec2helper(config = None):
        return XtrargonEc2Helper(config).load_credentials(section = 'ec2')

    # pylint: disable-msg=attribute-defined-outside-init
    def safe_init(self, options):
        self.lock_timeout = self.config['general']['lock_timeout']
        self.request_timeout = DEFAULT_REQUEST_TIMEOUT

        backups_status_uri = DEFAULT_BACKUPS_STATUS_URI
        forbid_vip_uri = DEFAULT_FORBID_VIP_URI

        self._clusters = {}

        if not isinstance(self.config.get('quorum'), dict):
            self.ec2helper = self._load_ec2helper()
        else:
            self.ec2helper = self._load_ec2helper(self.config['quorum'])

            if self.config['quorum'].get('backups_status_uri'):
                backups_status_uri = self.config['quorum']['backups_status_uri']

            if self.config['quorum'].get('forbid_vip_uri'):
                forbid_vip_uri = self.config['quorum']['forbid_vip_uri']

            if 'request_timeout' in self.config['quorum']:
                self.request_timeout = float(self.config['quorum']['request_timeout'])

            if self.config['quorum'].get('clusters'):
                self._clusters = self.config['quorum']['clusters']

        self.backups_status_uri = list(urisup.uri_help_split(backups_status_uri))
        self.forbid_vip_uri = list(urisup.uri_help_split(forbid_vip_uri))

    @staticmethod
    def _get_neighbor_node(host):
        node = host.lower()

        m = MATCH_DB_NODE(host)
        if not m:
            raise HttpReqErrJson(415, "invalid node host argument: %r" % node)

        letter = m.group(2)
        if letter not in DB_NODE_LETTERS:
            raise HttpReqErrJson(415, "invalid node host letter: %r" % node)

        host_splitted    = list(m.groups())
        host_splitted[1] = DB_NODE_LETTERS[letter]

        return ''.join(host_splitted)

    def _get_cluster_info(self, cluster):
        cluster = cluster.lower()
        info = deepcopy(self._clusters.get(cluster))

        if not info:
            return None

        info['host'] = cluster

        if cluster.get('replication_type') is None:
            info['replication_type'] = DB_REPLICATION_TYPE_DEFAULT
        elif cluster.get('replication_type') in DB_REPLICATION_TYPE_OPTS:
            info['replication_type'] = cluster.get('replication_type')
        else:
            info['replication_type'] = bool(cluster.get('replication_type'))

        return info

    def _build_uri(self, uri, host = None):
        uri = list(uri)
        if uri[1]:
            uri[1] = list(uri[1])
        else:
            uri[1] = [None, None, None, None]

        if host and not uri[1][2]:
            uri[1][2] = host

        if uri[1][3] == '0':
            uri[1][3] = self.config['general'].get('listen_port')

        return urisup.uri_help_unsplit(uri)

    def _get_routes_by_instance(self, cluster, client = None):
        r = {}

        if not client:
            client = self.ec2helper.boto3_client()

        instances = self.ec2helper.instance_info(
            client.describe_instances(
                Filters = [{'Name': 'tag:PrivateDnsName',
                            'Values': cluster['instances']}]))

        if not instances:
            LOG.error("unable to find any instance")
            return None

        for name, value in iteritems(instances):
            r[name] = {'id': value['id'],
                       'routes': {}}

            rtbs = client.describe_route_tables(
                Filters = [{'Name': 'route.instance-id',
                            'Values': value['id']}],
                RouteTableIds = cluster['routes'])

            if not rtbs or not rtbs.get('RouteTables'):
                LOG.info("unable to find a route for instance: %r", name)
                continue

            for rtb in rtbs['RouteTables']:
                for route in rtb['Routes']:
                    if route \
                       and route.get('DestinationCidrBlock') \
                       and route['DestinationCidrBlock'].starswith("%s/" % cluster['vip']):
                        if rtb['RouteTableId'] not in r[name]['routes']:
                            r[name]['routes'][rtb['RouteTableId']] = []
                        r[name]['routes'][rtb['RouteTableId']].append(route)

        return r


    def _prepare_replace_routes(self, cluster, client = None, replica = True):
        r = {'node_type': None,
             'instance_id': None,
             'instance_name': None,
             'routes': []}

        routes_instance = self._get_routes_by_instance(cluster, client)
        if not routes_instance:
            return None

        for name, value in iteritems(routes_instance):
            if not value['routes'] and replica:
                r['node_type'] = 'replica'
                r['instance_id'] = value['id']
                r['instance_name'] = name
                continue

            if value['routes'] and not replica:
                r['node_type'] = 'primary'
                r['instance_id'] = value['id']
                r['instance_name'] = name
                continue

            for route, info in iteritems(value):
                r['routes'].append({'DestinationCidrBlock': info['DestinationCidrBlock'],
                                    'RouteTableId': route})

        if not r['instance_id']:
            return False

        return r

    @staticmethod
    def _touch_freeze_vip(host):
        helpers.touch(DEFAULT_FREEZE_VIP_FILE_PATTERN % host)


    BACKUPS_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def backups_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.BACKUPS_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        host = self._get_neighbor_node(params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        req = None

        try:
            LOG.info("get backups status on %r", host)

            req = requests.get(self._build_uri(self.backups_status_uri, host),
                               timeout = self.request_timeout)
            if not req.text:
                return None

            try:
                return req.json()
            except ValueError:
                req.raise_for_status()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if req:
                req.close()
            self.LOCK.release()

        return None


    FORBID_VIP_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def forbid_vip_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.FORBID_VIP_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        host = self._get_neighbor_node(params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        req = None

        try:
            LOG.info("get forbid vip status on %r", host)

            req = requests.get(self._build_uri(self.forbid_vip_uri, host),
                               timeout = self.request_timeout)
            if not req.text:
                return None

            try:
                return req.json()
            except ValueError:
                req.raise_for_status()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if req:
                req.close()
            self.LOCK.release()

        return None


    REPLICA_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def replica_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.REPLICA_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        host = self._get_neighbor_node(params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        qdb = None

        try:
            LOG.info("get replica status on %r", host)
            qdb = QuorumDB()
            qdb.init_db(module = self, default_cnf_file = DEFAULT_CNF_FILE, host = host)
            qdb.create_connection()

            return qdb.get_slave_status()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if qdb:
                qdb.close_connection()
            self.LOCK.release()


    CHECK_CLUSTER_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)
    def check_cluster_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.VIP_CONNECTION_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % cluster['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        qdb = None

        try:
            client = self.ec2helper.boto3_client()

            r = self._prepare_replace_routes(cluster, client, replica = False)
            if r is None:
                raise HttpReqErrJson(404, "unable to find routes for cluster: %r" % cluster['host'])

            if cluster['replication_type'] == 'primary-replica':
                LOG.info("get primary status on %r", r['instance_name'])
                qdb = QuorumDB()
                qdb.init_db(module = self, default_cnf_file = DEFAULT_CNF_FILE, host = r['instance_name'])
                qdb.create_connection()

                if qdb.check_is_slave():
                    raise HttpReqErrJson(500, "SQL primary server is replica on %r" % r['instance_name'])

                return bool(qdb.check_is_master())

            if cluster['replication_type'] == 'primary-only':
                raise HttpReqErrJson(500, "SQL primary server can not be a replica on %r" % r['instance_name'])

            if cluster['replication_type'] == 'replica-only':
                raise HttpReqErrJson(500, "SQL replica server can not be a primary on %r" % r['instance_name'])

            return False
        except HttpReqErrJson:
            raise
        except Exception as e:
            raise HttpReqErrJson(503, repr(e))
        finally:
            if self.ec2helper:
                self.ec2helper.terminate()
            if qdb:
                qdb.close_connection()
            self.LOCK.release()


    FAILOVER_REPLICA_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def failover_replica(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.FAILOVER_REPLICA_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % cluster['host'])

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        qdb = None

        try:
            client = self.ec2helper.boto3_client()

            r = self._prepare_replace_routes(cluster, client)
            if r is None:
                raise HttpReqErrJson(404, "unable to find routes to failover for cluster: %r" % cluster['host'])

            if r is False:
                raise HttpReqErrJson(500, "unable to find instance to failover for cluster: %r" % cluster['host'])

            LOG.info("get replica status on %r", r['instance_name'])
            qdb = QuorumDB()
            qdb.init_db(module = self, default_cnf_file = DEFAULT_CNF_FILE, host = r['instance_name'])
            qdb.create_connection()

            if qdb.check_is_slave() != 'Yes':
                raise HttpReqErrJson(500, "SQL replica thread isn't running on %r" % r['instance_name'])

            if not qdb.replica_io_running():
                raise HttpReqErrJson(501, "SQL replica IO isn't running on %r" % r['instance_name'])

            if qdb.check_is_master():
                raise HttpReqErrJson(502, "SQL replica is also primary on %r" % r['instance_name'])

            for route in r['routes']:
                route['InstanceId'] = r['instance_id']
                client.replace_route(**route)

            self._touch_freeze_vip(cluster['host'])

            return True
        except HttpReqErrJson:
            raise
        except Exception as e:
            raise HttpReqErrJson(503, repr(e))
        finally:
            if self.ec2helper:
                self.ec2helper.terminate()
            if qdb:
                qdb.close_connection()
            self.LOCK.release()


    FREEZE_VIP_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def freeze_vip(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.VIP_CONNECTION_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % cluster['host'])

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            LOG.info("write freeze file for cluster %r", cluster['host'])
            self._touch_freeze_vip(cluster['host'])

            return True
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    VIP_CONNECTION_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def vip_connection_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.VIP_CONNECTION_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % cluster['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        qdb = None

        try:
            LOG.info("get VIP connection status on cluster %r", cluster['host'])
            qdb = QuorumDB()
            qdb.init_db(module = self, default_cnf_file = DEFAULT_CNF_FILE, host = cluster['vip'])
            qdb.create_connection()

            return qdb.test_connection()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if qdb:
                qdb.close_connection()
            self.LOCK.release()


    ROUTE_TABLES_FORCE_REPLICA_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)
    def route_tables_force_replica(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.ROUTE_TABLES_FORCE_REPLICA_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % cluster['host'])

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            client = self.ec2helper.boto3_client()

            r = self._prepare_replace_routes(cluster, client)
            if r is None:
                raise HttpReqErrJson(404, "unable to find routes to failover for cluster: %r" % cluster['host'])

            if r is False:
                raise HttpReqErrJson(500, "unable to find instance to failover for cluster: %r" % cluster['host'])

            for route in r['routes']:
                route['InstanceId'] = r['instance_id']
                client.replace_route(**route)

            self._touch_freeze_vip(cluster['host'])

            return r['routes']
        except HttpReqErrJson:
            raise
        except Exception as e:
            raise HttpReqErrJson(503, repr(e))
        finally:
            if self.ec2helper:
                self.ec2helper.terminate()
            self.LOCK.release()


    ROUTE_TABLES_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def route_tables_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.ROUTE_TABLES_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        cluster = self._get_cluster_info(params['host'])
        if not cluster:
            raise HttpReqErrJson(415, "invalid cluster name: %r" % cluster['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            r = self._get_routes_by_instance(cluster)
            if not r:
                raise HttpReqErrJson(404, "unable to find routes for cluster: %r" % cluster['host'])

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            raise HttpReqErrJson(503, repr(e))
        finally:
            if self.ec2helper:
                self.ec2helper.terminate()
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(QuorumModule())
    _start()
